//  This file is part of the Mussa source distribution.
//  http://mussa.caltech.edu/
//  Contact author: Tristan  De Buysscher, tristan@caltech.edu

// This program and all associated source code files are Copyright (C) 2005
// the California Institute of Technology, Pasadena, CA, 91125 USA.  It is
// under the GNU Public License; please see the included LICENSE.txt
// file for more information, or contact Tristan directly.


//  This file is part of the Mussa source distribution.
//  http://mussa.caltech.edu/
//  Contact author: Tristan  De Buysscher, tristan@caltech.edu

// This program and all associated source code files are Copyright (C) 2005
// the California Institute of Technology, Pasadena, CA, 91125 USA.  It is
// under the GNU Public License; please see the included LICENSE.txt
// file for more information, or contact Tristan directly.


//                        ----------------------------------------
//                           ---------- sequence.cc -----------
//                        ----------------------------------------
#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/operations.hpp>
namespace fs = boost::filesystem;

#include <boost/spirit/core.hpp>
#include <boost/spirit/actor/push_back_actor.hpp>
#include <boost/spirit/iterator/file_iterator.hpp>
#include <boost/spirit/utility/chset.hpp>
namespace spirit = boost::spirit;

#include "alg/sequence.hpp"
#include "mussa_exceptions.hpp"

#include <string>
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <set>

annot::annot() 
 : begin(0),
   end(0),
   type(""),
   name("")
{
}

annot::annot(int begin, int end, std::string type, std::string name)
 : begin(begin),
   end(end),
   type(type),
   name(name)
{
}

annot::~annot()
{
}

bool operator==(const annot& left, const annot& right)
{
  return ((left.begin== right.begin) and
          (left.end == right.end) and
          (left.type == right.type) and
          (left.name == right.name));
}

motif::motif(int begin, std::string motif)
 : annot(begin, begin+motif.size(), "motif", motif),
   sequence(motif)
{
}

motif::~motif()
{
}


Sequence::Sequence(AlphabetRef alphabet)
  : seq(new SeqSpan("", alphabet, SeqSpan::PlusStrand)),    
    motif_list(new MotifList)
{
}

Sequence::~Sequence()
{
}

Sequence::Sequence(const char *seq, AlphabetRef alphabet_, SeqSpan::strand_type strand_)
  : header(""),
    species(""),
    motif_list(new MotifList)
{
  set_filtered_sequence(seq, alphabet_, 0, npos, strand_);
}

Sequence::Sequence(const std::string& seq, 
                   AlphabetRef alphabet_,
                   SeqSpan::strand_type strand_) 
  : header(""),
    species(""),
    motif_list(new MotifList)
{
  set_filtered_sequence(seq, alphabet_, 0, seq.size(), strand_);
}

Sequence::Sequence(const Sequence& o)
  : seq(o.seq),
    header(o.header),
    species(o.species),
    annots(o.annots),
    motif_list(o.motif_list)
{
}

Sequence::Sequence(const Sequence* o)
  : seq(o->seq),
    header(o->header),
    species(o->species),
    annots(o->annots),
    motif_list(o->motif_list)
{
}

Sequence::Sequence(const SequenceRef o)
  : seq(new SeqSpan(o->seq)),
    header(o->header),
    species(o->species),
    annots(o->annots),
    motif_list(o->motif_list)
{
}

Sequence::Sequence(const SeqSpanRef& seq_ref)
  : seq(seq_ref),
    header(""),
    species(""),
    motif_list(new MotifList)
{
}

Sequence &Sequence::operator=(const Sequence& s)
{
  if (this != &s) {
    seq = s.seq;
    header = s.header;
    species = s.species;
    annots = s.annots;
    motif_list = s.motif_list;
  }
  return *this;
}

static void multiplatform_getline(std::istream& in, std::string& line)
{
  line.clear();
  char c;
  in.get(c);
  while(in.good() and !(c == '\012' or c == '\015') ) {
    line.push_back(c);
    in.get(c);
  }
  // if we have cr-lf eat it
  c = in.peek();
  if (c=='\012' or c == '\015') {
    in.get();
  }
}

void Sequence::load_fasta(fs::path file_path, int seq_num, int start_index, int end_index)
{
  load_fasta(file_path, reduced_nucleic_alphabet, seq_num, start_index, end_index);
}

//! load a fasta file into a sequence
void Sequence::load_fasta(fs::path file_path, AlphabetRef a, 
                          int seq_num, int start_index, int end_index)
{
  fs::fstream data_file;
  data_file.open(file_path, std::ios::in);

  if (!data_file.good())
  {    
    throw mussa_load_error("Sequence File: "+file_path.string()+" not found"); 
  } else {
    try {
      load_fasta(data_file, a, seq_num, start_index, end_index);
    } catch(sequence_empty_error e) {
      // there doesn't appear to be any sequence
      // catch and rethrow to include the filename
      std::stringstream msg;
      msg << "The selected sequence in " 
          << file_path.native_file_string()
          << " appears to be empty"; 
      throw sequence_empty_error(msg.str());
    } catch(sequence_empty_file_error e) {
      std::stringstream errormsg;
      errormsg << file_path.native_file_string()
               << " did not have any fasta sequences" << std::endl;
      throw sequence_empty_file_error(errormsg.str());
    } catch(sequence_invalid_load_error e) {
      std::ostringstream msg;
      msg << file_path.native_file_string();
      msg << " " << e.what();
      throw sequence_invalid_load_error(msg.str());
    }
  }
}

void Sequence::load_fasta(std::istream& file, 
                          int seq_num, int start_index, int end_index)
{
  load_fasta(file, reduced_nucleic_alphabet, seq_num, start_index, end_index);
}

void
Sequence::load_fasta(std::istream& data_file, AlphabetRef a, 
                     int seq_num, 
                     int start_index, int end_index)
{
  std::string file_data_line;
  int header_counter = 0;
  size_t line_counter = 0;
  bool read_seq = true;
  std::string rev_comp;
  std::string sequence_raw;
  std::string seq_tmp;      // holds sequence during basic filtering
  const Alphabet &alpha = Alphabet::get_alphabet(a);

  if (seq_num == 0) {
    throw mussa_load_error("fasta sequence number is 1 based (can't be 0)");
  }

  // search for the header of the fasta sequence we want
  while ( (!data_file.eof()) && (header_counter < seq_num) )
  {
    multiplatform_getline(data_file, file_data_line);
    ++line_counter;
    if (file_data_line.substr(0,1) == ">")
      header_counter++;
  }

  if (header_counter > 0) {
    header = file_data_line.substr(1);

    sequence_raw = "";

    while ( !data_file.eof() && read_seq ) {
      multiplatform_getline(data_file,file_data_line);
      ++line_counter;
      if (file_data_line.substr(0,1) == ">")
        read_seq = false;
      else {
        for (std::string::const_iterator line_i = file_data_line.begin();
             line_i != file_data_line.end();
             ++line_i)
         {
           if(alpha.exists(*line_i)) {
             sequence_raw += *line_i;
           } else {
            std::ostringstream msg;
            msg << "Unrecognized characters in fasta sequence at line ";
            msg << line_counter;
            throw sequence_invalid_load_error(msg.str());
           }
         }
      }
    }

    // Lastly, if subselection of the sequence was specified we keep cut out
    // and only keep that part
    // end_index = 0 means no end was specified, so cut to the end 
    if (end_index == 0)
      end_index = sequence_raw.size();

    // sequence filtering for upcasing agctn and convert non AGCTN to N
    if (end_index-start_index <= 0) {
      std::string msg("The selected sequence appears to be empty"); 
      throw sequence_empty_error(msg);
    }
    set_filtered_sequence(sequence_raw, a, start_index, end_index-start_index, SeqSpan::PlusStrand);
  } else {
    std::string errormsg("There were no fasta sequences");
    throw sequence_empty_file_error(errormsg);
  }
}

void Sequence::set_filtered_sequence(const std::string &in_seq,
                                     AlphabetRef alphabet_,
                                     size_type start,
                                     size_type count,
                                     SeqSpan::strand_type strand_)
{
  if ( count == npos)
    count = in_seq.size() - start;
  std::string new_seq;
  new_seq.reserve(count);

  // finally, the actual conversion loop
  const Alphabet& alpha_impl = Alphabet::get_alphabet(alphabet_); // go get one of our actual alphabets
  std::string::const_iterator seq_i = in_seq.begin()+start;
  for(size_type i = 0; i != count; ++i, ++seq_i)
  {
    if (alpha_impl.exists(*seq_i)) {
      new_seq.append(1, toupper(*seq_i));
    } else {
      new_seq.append(1, 'N');
    }
  }
  SeqSpanRef new_seq_ref(new SeqSpan(new_seq, alphabet_, strand_)); 
  seq = new_seq_ref;
}

void
Sequence::load_annot(fs::path file_path, int start_index, int end_index)
{
  if (not fs::exists(file_path)) {
    throw mussa_load_error("Annotation File " + file_path.string() + " was not found");
  }
  if (fs::is_directory(file_path)) {
    throw mussa_load_error(file_path.string() +
            " is a directory, please provide a file for annotations."
          );
  }    
  fs::fstream data_stream(file_path, std::ios::in);
  if (!data_stream)
  {
    throw mussa_load_error("Error loading annotation file " + file_path.string());
  }

  // so i should probably be passing the parse function some iterators
  // but the annotations files are (currently) small, so i think i can 
  // get away with loading the whole file into memory
  std::string data;
  char c;
  while(data_stream.good()) {
    data_stream.get(c);
    data.push_back(c);
  }
  data_stream.close();

  try {  
    parse_annot(data, start_index, end_index);
  } catch(annotation_load_error e) {
    std::ostringstream msg;
    msg << file_path.native_file_string()
        << " "
        << e.what();
    throw annotation_load_error(msg.str());
  }
}

/* If this works, yikes, this is some brain hurting code.
 *
 * what's going on is that when pb_annot is instantiated it stores references
 * to begin, end, name, type, declared in the parse function, then
 * when operator() is called it grabs values from those references
 * and uses that to instantiate an annot object and append that to our
 * annotation list.
 *
 * This weirdness is because the spirit library requires that actions
 * conform to a specific prototype operator()(IteratorT, IteratorT)
 * which doesn't provide any useful opportunity for me to actually
 * grab the results of our parsing.
 *
 * so I instantiate this structure in order to have a place to grab
 * my data from.
 */
  
struct push_back_annot {
  std::list<annot>& annot_list;
  int& begin;
  int& end;
  std::string& name;
  std::string& type;
  int &parsed;

  push_back_annot(std::list<annot>& annot_list_, 
                  int& begin_, 
                  int& end_, 
                  std::string& name_, 
                  std::string& type_,
                  int &parsed_) 
  : annot_list(annot_list_), 
    begin(begin_),
    end(end_),
    name(name_),
    type(type_),
    parsed(parsed_)
  {
  }

  void operator()(std::string::const_iterator, 
                  std::string::const_iterator) const 
  {
    //std::cout << "adding annot: " << begin << "|" << end << "|" << name << "|" << type << std::endl;
    annot_list.push_back(annot(begin, end, name, type));
    ++parsed;
  };
};

struct push_back_seq {
  std::list<Sequence>& seq_list;
  std::string& name;
  std::string& seq;
  int &parsed;

  push_back_seq(std::list<Sequence>& seq_list_,
                std::string& name_, 
                std::string& seq_,
                int &parsed_)
  : seq_list(seq_list_), 
    name(name_),
    seq(seq_),
    parsed(parsed_)
  {
  }

  void operator()(std::string::const_iterator, 
                  std::string::const_iterator) const 
  {
    // filter out newlines from our sequence
    std::string new_seq;
    for(std::string::const_iterator seq_i = seq.begin();
        seq_i != seq.end();
        ++seq_i)
    {
      if (*seq_i != '\015' && *seq_i != '\012') new_seq += *seq_i;
    }
    //std::cout << "adding seq: " << name << " " << new_seq << std::endl;
    
    Sequence s(new_seq);
    s.set_fasta_header(name);
    seq_list.push_back(s);
    ++parsed;
  };
};

void
Sequence::parse_annot(std::string data, int start_index, int end_index)
{
  int start=0;
  int end=0;
  std::string name;
  std::string type;
  std::string seq;
  std::list<annot> parsed_annots;
  std::list<Sequence> query_seqs;
  int parsed=0;

  bool ok = spirit::parse(data.begin(), data.end(),
              (
               //begin grammar
                 !(
                    (
                      spirit::alpha_p >> 
                      +(spirit::graph_p)
                    )[spirit::assign_a(species)] >> 
                    +(spirit::space_p)
                  ) >>
                  *(
                     ( // ignore html tags
                       *(spirit::space_p) >>
                       spirit::ch_p('<') >> 
                       +(~spirit::ch_p('>')) >>
                       spirit::ch_p('>') >>
                       *(spirit::space_p)
                     )
                   |
                    ( // parse an absolute location name
                     (spirit::uint_p[spirit::assign_a(start)] >> 
                      +spirit::space_p >>
                      spirit::uint_p[spirit::assign_a(end)] >> 
                      +spirit::space_p >>
                      ( 
                         spirit::alpha_p >> 
                         *spirit::graph_p
                      )[spirit::assign_a(name)] >> 
                      // optional type
                      !(
                          +spirit::space_p >>
                          (
                            spirit::alpha_p >>
                            *spirit::graph_p
                          )[spirit::assign_a(type)]
                      )
                      // to understand how this group gets set
                      // read the comment above struct push_back_annot
                     )[push_back_annot(parsed_annots, start, end, type, name, parsed)]
                   |
                    ((spirit::ch_p('>')|spirit::str_p("&gt;")) >> 
                       (*(spirit::print_p))[spirit::assign_a(name)] >>
                       spirit::eol_p >> 
                       (+(spirit::chset<>(Alphabet::nucleic_cstr)))[spirit::assign_a(seq)]
                     )[push_back_seq(query_seqs, name, seq, parsed)]
                    ) >>
                    *spirit::space_p
                   )
              //end grammar
              )).full;
  if (not ok) {
    std::stringstream msg;
    msg << "Error parsing annotation #" << parsed;
    throw annotation_load_error(msg.str());
  }
  // add newly parsed annotations to our sequence
  std::copy(parsed_annots.begin(), parsed_annots.end(), std::back_inserter(annots));
  // go seearch for query sequences 
  find_sequences(query_seqs.begin(), query_seqs.end());
}

void Sequence::add_annotation(const annot& a)
{
  annots.push_back(a);
}

const std::list<annot>& Sequence::annotations() const
{
  return annots;
}

void Sequence::copy_children(Sequence &new_seq, size_type start, size_type count) const
{
  new_seq.motif_list = motif_list;
  new_seq.annots.clear();

  for(std::list<annot>::const_iterator annot_i = annots.begin();
      annot_i != annots.end();
      ++annot_i)
  {
    size_type annot_begin= annot_i->begin;
    size_type annot_end = annot_i->end;

    if (annot_begin < start+count) {
      if (annot_begin >= start) {
        annot_begin -= start;
      } else {
        annot_begin = 0;
      }

      if (annot_end < start+count) {
        annot_end -= start;
      } else {
        annot_end = count;
      }

      annot new_annot(annot_begin, annot_end, annot_i->type, annot_i->name);
      new_seq.annots.push_back(new_annot);
    }
  }
}

Sequence
Sequence::subseq(size_type start, size_type count, SeqSpan::strand_type strand) const
{
  // FIXME: should i really allow a subsequence of an empty sequence?
  if (!seq) {
    Sequence new_seq;
    return new_seq;
  }

  Sequence new_seq = *this;
  new_seq.seq = seq->subseq(start, count, strand);
  if (seq->annotations()) {
    AnnotationsRef a(new Annotations(*(seq->annotations())));
    new_seq.seq->setAnnotations(a);
  }
  copy_children(new_seq, start, count);
  
  return new_seq;
}


// FIXME: This needs to be moved into SeqSpan
Sequence Sequence::rev_comp() const
{
  // a reverse complement is the whole opposite strand 
  return subseq(0, npos, SeqSpan::OppositeStrand);
}

const Alphabet& Sequence::get_alphabet() const
{
  return (seq) ? seq->get_alphabet() : Alphabet::empty_alphabet(); 
}

void Sequence::set_fasta_header(std::string header_)
{
  header = header_;
}

void Sequence::set_species(const std::string& name)
{
  species = name;
}

std::string Sequence::get_species() const
{
  return species;
}


std::string
Sequence::get_fasta_header() const
{
  return header;
}

std::string
Sequence::get_name() const
{
  if (header.size() > 0) 
    return header;
  else if (species.size() > 0)
    return species;
  else
    return "";
}

void Sequence::set_sequence(const std::string& s, AlphabetRef a) 
{
  set_filtered_sequence(s, a, 0, s.size(), SeqSpan::PlusStrand);
}

std::string Sequence::get_sequence() const
{
  return seq->sequence();
}

Sequence::const_reference Sequence::operator[](Sequence::size_type i) const
{
  return at(i);
}

void
Sequence::clear()
{
  seq.reset();
  header.clear();
  species.clear();
  annots.clear();
  motif_list.reset(new MotifList);
}

void
Sequence::save(fs::fstream &save_file)
{
  //fstream save_file;
  std::list<annot>::iterator annots_i;

  // not sure why, or if i'm doing something wrong, but can't seem to pass
  // file pointers down to this method from the mussa control class
  // so each call to save a sequence appends to the file started by mussa_class
  //save_file.open(save_file_path.c_str(), std::ios::app);

  save_file << "<Sequence>" << std::endl;
  save_file << *this << std::endl;
  save_file << "</Sequence>" << std::endl;

  save_file << "<Annotations>" << std::endl;
  save_file << species << std::endl;
  for (annots_i = annots.begin(); annots_i != annots.end(); ++annots_i)
  {
    save_file << annots_i->begin << " " << annots_i->end << " " ;
    save_file << annots_i->name << " " << annots_i->type << std::endl;
  }
  save_file << "</Annotations>" << std::endl;
  //save_file.close();
}

void
Sequence::load_museq(fs::path load_file_path, int seq_num)
{
  fs::fstream load_file;
  std::string file_data_line;
  int seq_counter;
  annot an_annot;
  std::string::size_type space_split_i;
  std::string annot_value;

  annots.clear();
  load_file.open(load_file_path, std::ios::in);

  seq_counter = 0;
  // search for the seq_num-th sequence 
  while ( (!load_file.eof()) && (seq_counter < seq_num) )
  {
    getline(load_file,file_data_line);
    if (file_data_line == "<Sequence>")
      seq_counter++;
  }
  getline(load_file, file_data_line);
  // looks like the sequence is written as a single line
  set_filtered_sequence(file_data_line, reduced_dna_alphabet, 0, file_data_line.size(), SeqSpan::PlusStrand);
  getline(load_file, file_data_line);
  getline(load_file, file_data_line);
  if (file_data_line == "<Annotations>")
  {
    getline(load_file, file_data_line);
    species = file_data_line;
    while ( (!load_file.eof())  && (file_data_line != "</Annotations>") )
    {
      getline(load_file,file_data_line);
      if ((file_data_line != "") && (file_data_line != "</Annotations>"))  
      {
        // need to get 4 values...almost same code 4 times...
        // get annot start index
        space_split_i = file_data_line.find(" ");
        annot_value = file_data_line.substr(0,space_split_i);
        an_annot.begin = atoi (annot_value.c_str());
        file_data_line = file_data_line.substr(space_split_i+1);
        // get annot end index
        space_split_i = file_data_line.find(" ");
        annot_value = file_data_line.substr(0,space_split_i);
        an_annot.end = atoi (annot_value.c_str());

        if (space_split_i == std::string::npos)  // no entry for type or name
        {
          std::cout << "seq, annots - no type or name\n";
          an_annot.type = "";
          an_annot.name = "";
        }
        else   // else get annot type
        {
          file_data_line = file_data_line.substr(space_split_i+1);
          space_split_i = file_data_line.find(" ");
          annot_value = file_data_line.substr(0,space_split_i);
          an_annot.type = annot_value;
          if (space_split_i == std::string::npos)  // no entry for name
          {
            std::cout << "seq, annots - no name\n";
            an_annot.name = "";
          }
          else          // get annot name
          {
            file_data_line = file_data_line.substr(space_split_i+1);
            space_split_i = file_data_line.find(" ");
            annot_value = file_data_line.substr(0,space_split_i);
            an_annot.type = annot_value;
          }
        }
        annots.push_back(an_annot);  // don't forget to actually add the annot
      }
      //std::cout << "seq, annots: " << an_annot.start << ", " << an_annot.end
      //     << "-->" << an_annot.type << "::" << an_annot.name << std::endl;
    }
  }
  load_file.close();
}


void Sequence::add_motif(const Sequence& a_motif)
{
  std::vector<int> motif_starts = find_motif(a_motif);

  for(std::vector<int>::iterator motif_start_i = motif_starts.begin();
      motif_start_i != motif_starts.end();
      ++motif_start_i)
  {
    motif_list->push_back(motif(*motif_start_i, a_motif.get_sequence()));
  }
}

void Sequence::clear_motifs()
{
  if (motif_list) 
    motif_list->clear();
  else 
    motif_list.reset(new MotifList);
}

const Sequence::MotifList& Sequence::motifs() const
{
  return *motif_list;
}

std::vector<int>
Sequence::find_motif(const Sequence& a_motif) const
{
  std::vector<int> motif_match_starts;
  Sequence norm_motif_rc;

  motif_match_starts.clear();
  // std::cout << "motif is: " << norm_motif << std::endl;

  if (a_motif.size() > 0)
  {
    //std::cout << "Sequence: none blank motif\n";
    motif_scan(a_motif, &motif_match_starts);

    norm_motif_rc = a_motif.rev_comp();;
    // make sure not to do search again if it is a palindrome
    if (norm_motif_rc != a_motif) {
      motif_scan(norm_motif_rc, &motif_match_starts);
    }
  }
  return motif_match_starts;
}

void
Sequence::motif_scan(const Sequence& a_motif, std::vector<int> * motif_match_starts) const
{
  // if there's no sequence we can't scan for it?
  // should this throw an exception?
  if (!seq) return;

  std::string::size_type seq_i = 0;
  Sequence::size_type motif_i = 0;
  Sequence::size_type motif_len = a_motif.length();
  Sequence::value_type motif_char;
  Sequence::value_type seq_char;

  while (seq_i < size())
  {
    // this is pretty much a straight translation of Nora's python code
    // to match iupac letter codes
    motif_char = toupper(a_motif[motif_i]);
    seq_char = toupper(seq->at(seq_i));
    if (motif_char =='N')
      motif_i++;
    else if (motif_char == seq_char)
      motif_i++;
    else if ((motif_char =='M') && ((seq_char=='A') || (seq_char=='C')))
      motif_i++;
    else if ((motif_char =='R') && ((seq_char=='A') || (seq_char=='G')))
      motif_i++;
    else if ((motif_char =='W') && ((seq_char=='A') || (seq_char=='T')))
      motif_i++;
    else if ((motif_char =='S') && ((seq_char=='C') || (seq_char=='G')))
      motif_i++;
    else if ((motif_char =='Y') && ((seq_char=='C') || (seq_char=='T')))
      motif_i++;
    else if ((motif_char =='K') && ((seq_char=='G') || (seq_char=='T')))
      motif_i++;
    else if ((motif_char =='V') && 
             ((seq_char=='A') || (seq_char=='C') || (seq_char=='G')))
      motif_i++;
    else if ((motif_char =='H') && 
             ((seq_char=='A') || (seq_char=='C') || (seq_char=='T')))
      motif_i++;
    else if ((motif_char =='D') &&
             ((seq_char=='A') || (seq_char=='G') || (seq_char=='T')))
      motif_i++;
    else if ((motif_char =='B') &&
             ((seq_char=='C') || (seq_char=='G') || (seq_char=='T')))
      motif_i++;
    else
    {
      // if a motif doesn't match, erase our current trial and try again
      seq_i -= motif_i;
      motif_i = 0;
    }

    // end Nora stuff, now we see if a match is found this pass
    if (motif_i == motif_len)
    {
      annot new_motif;
      motif_match_starts->push_back(seq_i - motif_len + 1);
      motif_i = 0;
    }

    seq_i++;
  }
  //std::cout << std::endl;
}

void Sequence::add_string_annotation(std::string a_seq, 
                                     std::string name)
{
  std::vector<int> seq_starts = find_motif(a_seq);
  
  //std::cout << "searching for " << a_seq << " found " << seq_starts.size() << std::endl;

  for(std::vector<int>::iterator seq_start_i = seq_starts.begin();
      seq_start_i != seq_starts.end();
      ++seq_start_i)
  {
    annots.push_back(annot(*seq_start_i, 
                           *seq_start_i+a_seq.size(),
                           "",
                           name));
  }
}

void Sequence::find_sequences(std::list<Sequence>::iterator start, 
                              std::list<Sequence>::iterator end)
{
  while (start != end) {
    add_string_annotation(start->get_sequence(), start->get_fasta_header());
    ++start;
  }
}


std::ostream& operator<<(std::ostream& out, const Sequence& s)
{
  if (s.seq) {
    for(Sequence::const_iterator s_i = s.begin(); s_i != s.end(); ++s_i) {
      out << *s_i;
    }
  }
  return out;
}

bool operator<(const Sequence& x, const Sequence& y)
{
  Sequence::const_iterator x_i = x.begin();
  Sequence::const_iterator y_i = y.begin();
  // for sequences there's some computation associated with computing .end
  // so lets cache it.
  Sequence::const_iterator xend = x.end();
  Sequence::const_iterator yend = y.end();
  while(1) {
    if( x_i == xend and y_i == yend ) {
      return false;
    } else if ( x_i == xend ) {
      return true;
    } else if ( y_i == yend ) {
      return false;
    } else if ( (*x_i) < (*y_i)) {
      return true;
    } else if ( (*x_i) > (*y_i) ) {
      return false;
    } else {
      ++x_i;
      ++y_i;
    }
  }
}

template <typename Iter1, typename Iter2>
static
bool sequence_insensitive_equality(Iter1 abegin, Iter1 aend, Iter2 bbegin, Iter2 bend)
{
  Iter1 aseq_i = abegin;
  Iter2 bseq_i = bbegin;
  if (aend-abegin == bend-bbegin) {
    // since the length of the two sequences is equal, we only need to
    // test one.
    for(; aseq_i != aend; ++aseq_i, ++bseq_i) {
      if (toupper(*aseq_i) != toupper(*bseq_i)) {
        return false;
      }
    }
    return true;
  } else {
    return false;
  }
}

bool operator==(const Sequence& x, const Sequence& y) 
{
  if (x.seq and y.seq) {
    // both x and y are defined
    if (SeqSpan::isFamily(x.seq, y.seq)) {
      // both are part of the same SeqSpan tree
      return *(x.seq) == *(y.seq);
    } else {
      // we'll have to do a real comparison
      return sequence_insensitive_equality<SeqSpan::const_iterator, SeqSpan::const_iterator>(
               x.begin(), x.end(),
               y.begin(), y.end()
             );
    }
  } else {
    // true if they're both empty (with either a null SeqSpanRef or 
    // a zero length string
    return (x.size() == y.size());
  }
}

bool operator!=(const Sequence& x, const Sequence& y) 
{
  return not operator==(x, y);
}

